
from os_spage.validator import check_error_reason, simple_check_url


def test_check_error_reason():
    valid_data = ['HTTP 404', 'RULE 16', 'DNS -2', 'SERVER 110', 'SSL -2']
    for data in valid_data:
        assert check_error_reason(data) == True

    invalid_data = ['HTTP404', 'RULE  16', 'RULE TEST']
    for data in invalid_data:
        assert check_error_reason(data) == False


def test_simple_check_url():
    valid_data = [
        "http://www.google.com/",
        "https://www.google.com/",
    ]
    for data in valid_data:
        assert simple_check_url(data) == True

    invalid_data = [
        "htp://www.google.com/",
        "https/www.google.com/",
    ]
    for data in invalid_data:
        assert simple_check_url(data) == False


import pkgutil
import sys
from .spage_reader import SpageReader, read as read_spage
from .spage_writer import SpageWriter, write
from .offpage_reader import OffpageReader, read as read_offpage


def __not_supported_mode(name, **kwargs):
    raise ValueError("mode must be 'r' or 'w'")


def __not_supported_page_type(name, **kwargs):
    raise ValueError("page_type must be 'spage' or 'offpage'")


def read(s, page_type='spage'):
    r = {'spage': read_spage, 'offpage': read_offpage}.get(
        page_type, __not_supported_page_type)
    return r(s)


def open_file(name, mode, **kwargs):
    r = {'w': SpageWriter,
         'r': {'spage': SpageReader,
               'offpage': OffpageReader}}.get(mode, __not_supported_mode)
    if mode == 'r':
        r = r.get(kwargs.pop('page_type', 'spage'), __not_supported_page_type)

    return r(name, **kwargs)


__all__ = ['__version__', 'version_info', 'open_file']

__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
version_info = tuple(int(v) if v.isdigit() else v
                     for v in __version__.split('.'))

if sys.version_info < (2, 7):
    sys.exit("os-spage %s requires Python 2.7+" % __version__)

del pkgutil



from collections import OrderedDict
from datetime import datetime


class RecordTypes(object):
    FLAT       = 'flat'
    DELETED    = 'deleted'
    COMPRESSED = 'compressed'


class InnerHeaderKeys(object):
    VERSION           = 'Version'
    TYPE              = 'Type'
    FETCH_TIME        = 'Fetch-Time'
    ORIGINAL_SIZE     = 'Original-Size'
    STORE_SIZE        = 'Store-Size'
    BATCH_ID          = 'batchID'
    ATTACH            = 'attach'
    IP_ADDRESS        = 'IP-Address'
    SPIDER_ADDRESS    = 'Spider-Address'
    DIGEST            = 'Digest'
    USER_AGENT        = 'User-Agent'
    FETCH_IP          = 'Fetch-IP'
    NODE_FETCH_TIME   = 'Node-Fetch-Time'
    ERROR_REASON      = 'Error-Reason'


INNER_HEADER_SCHEMA = {
    "type": "object",
    "properties": OrderedDict([
        (InnerHeaderKeys.VERSION, {
            "type": "string",
            "default": "1.2",
        }),
        (InnerHeaderKeys.TYPE, {  # autofill or specify
            "type": "string",
            "enum": set([getattr(RecordTypes, i) for i in dir(RecordTypes) if not i.startswith('_')]) ,
        }),
        (InnerHeaderKeys.FETCH_TIME, {  # record store time
            "anyOf": [
                {
                    "type": "datetime",
                },
                {
                    "type": "string",
                    "format": "readable_time"
                },
            ],
            "default": datetime.now,
        }),
        (InnerHeaderKeys.ORIGINAL_SIZE, {  # data(html) size, autofill
            "type": "number",
        }),
        (InnerHeaderKeys.STORE_SIZE, {  # store size, autofill
            "type": "number",
        }),
        (InnerHeaderKeys.BATCH_ID, {  # batch identity
            "type": "string",
            "minLength": 3,
            "default": '__CHANGE_ME__',
        }),
        (InnerHeaderKeys.ATTACH, {
            "type": "string",
        }),
        (InnerHeaderKeys.IP_ADDRESS, {  # remote host ip
            "type": "string",
            "format": "ipv4",
            "default": "0.0.0.0",
        }),
        (InnerHeaderKeys.SPIDER_ADDRESS, {  # spider node identity
            "type": "string",
            "default": "0.0.0.0",
        }),
        (InnerHeaderKeys.DIGEST, {  # can be html md5
            "type": "string",
            "default": "0" * 32,
            "maxLength": 32,
            "minLength": 32,
        }),
        (InnerHeaderKeys.USER_AGENT, {
            "type": "string",
        }),
        (InnerHeaderKeys.FETCH_IP, {  # generate page machine ip
            "type": "string",
            "format": "ipv4",
            "default": "0.0.0.0",
        }),
        (InnerHeaderKeys.NODE_FETCH_TIME, {  # real fetch time
            "anyOf": [
                {
                    "type": "datetime",
                },
                {
                    "type": "string",
                    "format": "readable_time"
                },
            ]
        }),
        (InnerHeaderKeys.ERROR_REASON, {
            "type": "string",
            "format": "error_reaseon",
        }),
    ]),
}

class SpageKeys(object):
    URL          = 'url'
    INNER_HEADER = 'inner_header'
    HTTP_HEADER  = 'http_header'
    DATA         = 'data'


META_SCHEMA = {
    "type": "object",
    "properties": {
        SpageKeys.URL: {
            "type": "string",
            "format": "url",
        },
        SpageKeys.INNER_HEADER: INNER_HEADER_SCHEMA,
        SpageKeys.HTTP_HEADER: {
            "type": "object"
        },
        SpageKeys.DATA: {
            "type": "bytes",
        },
    },
    "required": [SpageKeys.URL, SpageKeys.INNER_HEADER],
}



import zlib

import pytest
from jsonschema import ValidationError
from os_spage import open_file, read, write
from os_spage.compat import iteritems
from os_spage.default_schema import InnerHeaderKeys as I_KEYS
from os_spage.default_schema import RecordTypes as R_TYPES
from os_spage.default_schema import SpageKeys as S_KEYS


def test_write_invalid_data(tmpdir):
    with tmpdir.as_cwd():
        f = open_file('test', 'w')
        with pytest.raises(ValidationError):
            f.write(url='abc', inner_header={I_KEYS.BATCH_ID: 'test'})


def check_inner_header(w_inner_header, r_inner_header):
    if not w_inner_header:
        assert r_inner_header[I_KEYS.BATCH_ID] == '__CHANGE_ME__'
        return

    for k, v in iteritems(w_inner_header):
        k = str(k)
        if k != I_KEYS.STORE_SIZE:
            assert r_inner_header[k] == str(v)


def check_http_header(w_http_header, r_http_header):
    if w_http_header is None:
        assert r_http_header == {}
        return
    for k, v in iteritems(w_http_header):
        assert r_http_header[str(k)] == str(v)


def check_data(w_data, r_data, r_type):
    if w_data is None:
        assert r_data is None
        return
    if r_type == R_TYPES.COMPRESSED:
        assert zlib.decompress(r_data) == w_data
    else:
        assert r_data == w_data


RECORDS = [
    # inner_header, http_header, data
    (None, None, None),
    ({I_KEYS.BATCH_ID: 'test'}, {'k1': 'v1'}, b'hello'),
    ({I_KEYS.BATCH_ID: 'test'}, {}, b'hello'),
    ({I_KEYS.BATCH_ID: 'test'}, {'k1': 'v1'}, None),
    ({}, {'k1': 'v1'}, b'hello'),
    ({}, {'k1': 'v1'}, None),
]


def test_spage_reader_and_writer(tmpdir):
    base_url = "http://www.test.com/"
    filename_prefix = 'test_file_'
    with tmpdir.as_cwd():
        f = open_file(filename_prefix, 'w', roll_size=100)
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            f.write(url, inner_header=inner_header,
                    http_header=http_header, data=data, flush=True)
        f.close()
        f = open_file(filename_prefix, 'r')
        idx = 0
        for record in f.read():
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_data(data, record[S_KEYS.DATA],
                       record[S_KEYS.INNER_HEADER][I_KEYS.TYPE])


def test_genergal_read_and_write(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename = 'test_file'
        f = open(filename, 'wb')
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            write(f, url, inner_header=inner_header,
                  http_header=http_header, data=data)
        f.close()
        f = open(filename, 'rb')
        idx = 0
        for record in read(f):
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_data(data, record[S_KEYS.DATA],
                       record[S_KEYS.INNER_HEADER][I_KEYS.TYPE])


CUSTOM_RECORDS = [
    # inner_header, http_header, data
    ({I_KEYS.BATCH_ID: 'test', I_KEYS.TYPE: R_TYPES.COMPRESSED,
      I_KEYS.ORIGINAL_SIZE: 5}, {'k1': 'v1'}, zlib.compress(b'hello')),
    ({I_KEYS.BATCH_ID: 'test', I_KEYS.TYPE: R_TYPES.FLAT,
      I_KEYS.ORIGINAL_SIZE: 5}, {'k1': 'v1'}, b'hello'),
    ({I_KEYS.BATCH_ID: 'test', I_KEYS.TYPE: R_TYPES.DELETED,
      I_KEYS.ORIGINAL_SIZE: 5}, {'k1': 'v1'}, b'hello'),
]


def check_custom_data(w_data, r_data):
    if w_data is None:
        assert r_data is None
        return
    assert r_data == w_data


def test_write_custom_data_and_read(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename_prefix = 'test_file_'
        f = open_file(filename_prefix, 'w', roll_size=100)
        idx = 0
        for inner_header, http_header, data in CUSTOM_RECORDS:
            url = base_url + str(idx)
            idx += 1
            f.write(url, inner_header=inner_header,
                    http_header=http_header, data=data)
        f.close()
        f = open_file(filename_prefix, 'r')
        idx = 0
        for record in f.read():
            inner_header, http_header, data = CUSTOM_RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record[S_KEYS.URL]
            check_inner_header(inner_header, record[S_KEYS.INNER_HEADER])
            check_http_header(http_header, record[S_KEYS.HTTP_HEADER])
            check_custom_data(data, record[S_KEYS.DATA])


INVALID_CUSTOM_RECORDS = [
    # inner_header, http_header, data
    ({I_KEYS.BATCH_ID: 'test', I_KEYS.TYPE: R_TYPES.COMPRESSED},
     {'k1': 'v1'}, zlib.compress(b'hello')),
    ({I_KEYS.BATCH_ID: 'test', I_KEYS.ORIGINAL_SIZE: 5},
     {'k1': 'v1'}, zlib.compress(b'hello')),
]


def test_write_invalid_custom_data(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename_prefix = 'test_file_'
        f = open_file(filename_prefix, 'w', roll_size=100)
        idx = 0
        for inner_header, http_header, data in INVALID_CUSTOM_RECORDS:
            url = base_url + str(idx)
            idx += 1
            with pytest.raises(ValueError):
                f.write(url, inner_header=inner_header,
                        http_header=http_header, data=data)
        f.close()



import sys
import operator

PY3 = sys.version_info[0] >= 3

if PY3:
    from io import StringIO
    iteritems = operator.methodcaller("items")

else:
    from StringIO import StringIO
    iteritems = operator.methodcaller("iteritems")
